/*
 * Copyright (C) 2020-2022, IOPSYS Software Solutions AB.
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include "vendor_uspd.h"
#include "os_utils.h"
#include "common_defs.h"
#include "data_model.h"
#include "str_vector.h"
#include "json.h"
#include "text_utils.h"
#include "uptime.h"

#include <uci.h>
#include <libubus.h>
#include <libubox/blob.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>

// Timeout in milliseconds
#define USP_OBJECT_NAME "usp.raw"
#define USP_PROTO "usp"
#define DEFAULT_IPC_TIMEOUT 10000  // 10 Sec
#define MIN_NUM_TO_GROUP (10)
#define NUMENTRY "NumberOfEntries"
#define MAX_CACHE_TIME (30)
#define MAX_GROUP_SEP 2
#define ROOT_PATH "Device."
#define USPD_EVENT "usp.event"
#define USPD_ADD_OBJ_EV_NAME "usp.AddObj"
#define USPD_DEL_OBJ_EV_NAME "usp.DelObj"

typedef enum {
	DMT_STRING,
	DMT_UNINT,
	DMT_INT,
	DMT_UNLONG,
	DMT_LONG,
	DMT_BOOL,
	DMT_TIME,
	DMT_HEXBIN,
	DMT_BASE64,
	DMT_COMMAND,
	DMT_EVENT,
	DMT_OBJECT,
} dmt_type_t;

const char *DMT_STR_MAP[] = {
	[DMT_STRING] = "xsd:string",
	[DMT_UNINT] = "xsd:unsignedInt",
	[DMT_INT] = "xsd:int",
	[DMT_UNLONG] = "xsd:unsignedLong",
	[DMT_LONG] = "xsd:long",
	[DMT_BOOL] = "xsd:boolean",
	[DMT_TIME] = "xsd:dateTime",
	[DMT_HEXBIN] = "xsd:hexBinary",
	[DMT_BASE64] = "xsd:base64",
	[DMT_COMMAND] = "xsd:command",
	[DMT_EVENT] = "xsd:event",
	[DMT_OBJECT] = "xsd:object",
};

static char *CONST_DM[] = {
	"Device.DeviceInfo.SoftwareVersion",
	"Device.DeviceInfo.ProductClass",
	"Device.DeviceInfo.Manufacturer",
	"Device.DeviceInfo.ModelName",
	"Device.DeviceInfo.HardwareVersion",
	"Device.DeviceInfo.ManufacturerOUI",
	"Device.DeviceInfo.SerialNumber",
};

typedef enum {
	USPD_INIT,
	USPD_PRE_START,
	USPD_START
} uspd_state_t;

struct uspd_global {
	struct ubus_context *ubus_ctx; // ubus context
	struct ubus_event_handler ubus_notify_event; // ubus usp.event to notify about events
	struct ubus_event_handler ubus_add_event;    // ubus event to update add_instances
	struct ubus_event_handler ubus_del_event;    // ubus event to delete instance
	int mq_sockets[2];
	str_vector_t group_vec;
	kv_vector_t cached_kv;
	str_vector_t cached_exclude_vec;
	int cached_index;
	int last_gid;
	int cached_interval;
	int ipc_timeout;
	unsigned last_cached_time;
	int trans_id;
	pthread_cond_t ubus_cond;
	pthread_mutex_t ubus_lock;
	uspd_state_t op_state;
};

static struct uspd_global g_uspd;
#define mq_rx_socket  g_uspd.mq_sockets[0]
#define mq_tx_socket  g_uspd.mq_sockets[1]
#define IS_UBUS_Q_READY ((mq_tx_socket == -1) ? false : true)
#define STR(s) #s

typedef enum
{
	CMD_NULL,
	CMD_GROUP_GET,
	CMD_GROUP_SET,
	CMD_ADD,
	CMD_DEL,
	CMD_GET,
	CMD_SET,
	CMD_OPERATE_SYNC,
	CMD_OPERATE_ASYNC,
	CMD_INSTANCES,
	CMD_TRAN_START,
	CMD_TRAN_COMMIT,
	CMD_TRAN_ABORT,
} uspd_cmd_t;

typedef struct
{
	uspd_cmd_t cmd;
	int len;
	int inst;
	int fault;
	char *path;
	char *p_value;
	str_vector_t *vec;
	kv_vector_t *kv;
	kv_vector_t *kv_out;
	unsigned *p_types;
	int *p_instance;
	int *p_failure_index;
} vendor_data_t;

// Forward declarations
extern int CountPathSeparator(char *path);
extern dm_node_t *FindNodeFromHash(dm_hash_t hash);

static void register_operates_events(struct blob_attr *param_blob, dmt_type_t bbf_type);
static void update_adaptive_cached_time();
static void dont_cached_params(str_vector_t *cached_vec);
static char **blob_array_str_list(struct blob_attr *attr, int *list_nr);
static dmt_type_t get_dmt_type(char *type);
static bool get_bool_from_str(char *bool_str);

// Private functions
static char *stringify_cmd(int cmd)
{
	switch(cmd) {
	case CMD_NULL:
		return STR(CMD_NULL);
	case CMD_GROUP_GET:
		return STR(CMD_GROUP_GET);
	case CMD_GROUP_SET:
		return STR(CMD_GROUP_SET);
	case CMD_ADD:
		return STR(CMD_ADD);
	case CMD_DEL:
		return STR(CMD_DEL);
	case CMD_GET:
		return STR(CMD_GET);
	case CMD_SET:
		return STR(CMD_SET);
	case CMD_OPERATE_SYNC:
		return STR(CMD_OPERATE_SYNC);
	case CMD_OPERATE_ASYNC:
		return STR(CMD_OPERATE_ASYNC);
	case CMD_INSTANCES:
		return STR(CMD_INSTANCES);
	case CMD_TRAN_START:
		return STR(CMD_TRAN_START);
	case CMD_TRAN_COMMIT:
		return STR(CMD_TRAN_COMMIT);
	case CMD_TRAN_ABORT:
		return STR(CMD_TRAN_ABORT);
	}

	return NULL;
}

static dm_node_t *dm_node_present(char *path)
{
	dm_hash_t hash;
	int err;

	err = DM_PRIV_CalcHashFromPath(path, NULL, &hash);
	if (err != USP_ERR_OK)
	{
		return NULL;
	}

	return FindNodeFromHash(hash);
}

static int get_value_from_uci(char *path, char *value, size_t max_value_len)
{
	struct uci_context *uci_ctx;
	struct uci_ptr ptr;
	int ret = 0;

	if (!path || !value || max_value_len == 0)
		return -1;

	uci_ctx = uci_alloc_context();
	if (!uci_ctx)
		return -1;

	if (uci_lookup_ptr(uci_ctx, &ptr, path, true) != UCI_OK) {
		ret = -1;
		goto exit;
	}

	if ((ptr.flags & UCI_LOOKUP_COMPLETE)
		&& (ptr.o != NULL)
		&& (ptr.o->v.string!=NULL)) {
		ret = 0;
		USP_STRNCPY(value, ptr.o->v.string, max_value_len);
	}

exit:
	uci_free_context(uci_ctx);
	return ret;
}

static int uspd_call(struct ubus_context *ctx, char *method,
		     struct blob_buf *data, ubus_data_handler_t callback,
		     vendor_data_t *arg)
{
	uint32_t id;
	int fault = UBUS_STATUS_OK;

	fault = ubus_lookup_id(ctx, USP_OBJECT_NAME, &id);
	if (fault) {
		USP_LOG_Error("[%s:%d] %s not present (%s)",
			      __func__, __LINE__, USP_OBJECT_NAME, ubus_strerror(fault));
		if (arg)
			arg->fault = USP_ERR_INTERNAL_ERROR;
		return USP_ERR_INTERNAL_ERROR;
	}

	// Invoke Ubus to get data from uspd
	fault = ubus_invoke(ctx, id, method, data->head, callback, arg, g_uspd.ipc_timeout);
	if (fault) {
		USP_LOG_Error("[%s:%d] ubus call for (%s)method failed(%s)",
			      __func__, __LINE__, method, ubus_strerror(fault));
		if (arg)
			arg->fault = USP_ERR_INTERNAL_ERROR;
		return USP_ERR_INTERNAL_ERROR;
	}

	return USP_ERR_OK;
}

static int uspd_call_inline_nocb(char *method, struct blob_buf *bb)
{
	struct ubus_context *ctx = ubus_connect(NULL);
	if (!ctx) {
		USP_LOG_Error("[%s:%d] ubus_connect failed", __func__, __LINE__);
		return USP_ERR_INTERNAL_ERROR;
	}

	uspd_call(ctx, method, bb, NULL, NULL);

	ubus_free(ctx);
	return USP_ERR_OK;
}

static void uspd_ubus_complete_cb(struct ubus_request *req, int ret)
{
    USP_FREE(req);
}

static int uspd_call_async(char *method, struct blob_buf *data, ubus_data_handler_t callback, void *priv)
{
	uint32_t id;
	int fault = UBUS_STATUS_OK;
	struct ubus_request *req;

	fault = ubus_lookup_id(g_uspd.ubus_ctx, USP_OBJECT_NAME, &id);
	if (fault) {
		USP_LOG_Error("[%s:%d] %s not present (%s)",
			      __func__, __LINE__, USP_OBJECT_NAME, ubus_strerror(fault));
		return USP_ERR_INTERNAL_ERROR;
	}

	req = USP_MALLOC(sizeof(struct ubus_request));
	if (req == NULL) {
		USP_LOG_Error("Out of memory!");
		return USP_ERR_RESOURCES_EXCEEDED;
	}

	memset(req, 0, sizeof(struct ubus_request));
	fault = ubus_invoke_async(g_uspd.ubus_ctx, id, method, data->head, req);
	if (fault) {
		USP_LOG_Error("[%s:%d] ubus async call failed(%s)",
			      __func__, __LINE__, ubus_strerror(fault));
		USP_FREE(req);
		return USP_ERR_INTERNAL_ERROR;
	}

	req->data_cb = callback;
	req->complete_cb = uspd_ubus_complete_cb;
	req->priv = priv;

	ubus_complete_request_async(g_uspd.ubus_ctx, req);

	return USP_ERR_OK;
}

static struct blob_attr *get_parameters(struct blob_attr *msg)
{
	struct blob_attr *params = NULL;
	struct blob_attr *cur;
	int rem;

	blobmsg_for_each_attr(cur, msg, rem) {
		if (blobmsg_type(cur) == BLOBMSG_TYPE_ARRAY) {
			params = cur;
			break;
		}
	}
	return params;
}

static int detect_fault(struct blob_attr *msg)
{
	int fault = USP_ERR_OK;
	enum {
		FAULT_ID,
		__FAULT_MAX
	};
	struct blob_attr *f_attr[__FAULT_MAX] = {NULL};
	const struct blobmsg_policy fp[__FAULT_MAX] = {
		{ "fault", BLOBMSG_TYPE_INT32 }
	};

	blobmsg_parse(fp, __FAULT_MAX, f_attr, blobmsg_data(msg),
			    blobmsg_len(msg));
	if (f_attr[FAULT_ID])
		fault = blobmsg_get_u32(f_attr[FAULT_ID]);

	return fault;
}

static void get_parameters_in_kv(struct blob_attr *params, kv_vector_t *kv_out)
{
	struct blob_attr *cur;
	size_t rem;
	enum {
		P_PARAM,
		P_VALUE,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "parameter", BLOBMSG_TYPE_STRING },
		{ "value", BLOBMSG_TYPE_STRING },
	};

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_PARAM] && tb[P_VALUE]) {
			USP_ARG_Add(kv_out,
				    blobmsg_get_string(tb[P_PARAM]),
				    blobmsg_get_string(tb[P_VALUE]));
		}
	}

}

static void _sync_operate_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	struct blob_attr *results, *rtemp;
	kv_vector_t *kv_out;
	size_t rem;

	if (!msg)
		return;

	kv_out = arg->kv_out;

	results = get_parameters(msg);
	if (results == NULL)
		return;

	blobmsg_for_each_attr(rtemp, results, rem) {
		arg->fault = detect_fault(rtemp);
		if (arg->fault != USP_ERR_OK)
			return;

		struct blob_attr *params = get_parameters(rtemp);

		if (params == NULL)
			return;

		get_parameters_in_kv(params, kv_out);
		break;
	}
}

static void _add_object_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	size_t rem;
	struct blob_attr *results, *cur;
	vendor_data_t *arg = req->priv;
	enum {
		P_INSTANCE,
		P_FAULT,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "instance", BLOBMSG_TYPE_STRING },
		{ "fault", BLOBMSG_TYPE_INT32 },
	};

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	results = get_parameters(msg);
	if (results == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_for_each_attr(cur, results, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL, NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_FAULT])
			arg->fault = blobmsg_get_u32(tb[P_FAULT]);

		if (tb[P_INSTANCE])
			*arg->p_instance = atoi(blobmsg_get_string(tb[P_INSTANCE]));
		break;
	}

}

static void _set_value_single_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	arg->fault = detect_fault(msg);
}

static void _set_value_group_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	vendor_data_t *arg = req->priv;
	enum {
		P_PATH,
		P_FAULT,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "path", BLOBMSG_TYPE_STRING },
		{ "fault", BLOBMSG_TYPE_INT32 }
	};

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	params = get_parameters(msg);
	if (params == NULL) {
		return;
	}

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_FAULT] && tb[P_PATH]) {
			char *path = blobmsg_get_string(tb[P_PATH]);

			arg->fault = blobmsg_get_u32(tb[P_FAULT]);
			*arg->p_failure_index = KV_VECTOR_FindKey(arg->kv, path, 0);
			break;
		}
	}
}

static void _del_object_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	vendor_data_t *arg = req->priv;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	params = get_parameters(msg);
	if (params == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_for_each_attr(cur, params, rem) {
		enum {
			P_FAULT,
			__P_MAX
		};
		const struct blobmsg_policy p[__P_MAX] = {
			{ "fault", BLOBMSG_TYPE_INT32 }
		};
		struct blob_attr *tb[__P_MAX] = {NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_FAULT]) {
			arg->fault = blobmsg_get_u32(tb[P_FAULT]);
			break;
		}
	}
}

static int _set_value_single(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "value", arg->p_value);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_uspd.trans_id);

	uspd_call(g_uspd.ubus_ctx, "set", &bb, _set_value_single_cb, arg);

	blob_buf_free(&bb);
	return USP_ERR_OK;
}

static int _set_value_group(vendor_data_t *arg)
{
	struct blob_buf bb = {};
	void *array;
	int i;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	array = blobmsg_open_array(&bb, "pv_tuple");
	for (i = 0; i < arg->kv->num_entries; i++) {
		void *table = blobmsg_open_table(&bb, NULL);

		blobmsg_add_string(&bb, "path", arg->kv->vector[i].key);
		blobmsg_add_string(&bb, "value", arg->kv->vector[i].value);
		blobmsg_close_table(&bb, table);
	}
	blobmsg_close_array(&bb, array);

	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_uspd.trans_id);

	uspd_call(g_uspd.ubus_ctx, "setm_values", &bb, _set_value_group_cb, arg);

	blob_buf_free(&bb);
	return USP_ERR_OK;
}

static int _add_object(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_uspd.trans_id);

	uspd_call(g_uspd.ubus_ctx, "add_object", &bb, _add_object_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static int _del_object(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_uspd.trans_id);

	uspd_call(g_uspd.ubus_ctx, "del_object", &bb, _del_object_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static int _operate_sync_handler(vendor_data_t *arg)
{
	struct blob_buf bb = {};
	char path[MAX_DM_PATH] = {'\0'}, *action;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	if (arg->kv_out)
		KV_VECTOR_Init(arg->kv_out);

	// separate path and command
	action = TEXT_UTILS_SplitPath(arg->path, path, MAX_DM_PATH);
	if (arg->kv != NULL) {
		void *table = blobmsg_open_table(&bb, "input");

		for (int i = 0; i < arg->kv->num_entries; ++i) {
			kv_pair_t *kv;

			kv = &arg->kv->vector[i];
			blobmsg_add_string(&bb, kv->key, kv->value);
		}
		blobmsg_close_table(&bb, table);
	}

	blobmsg_add_string(&bb, "path", path);
	blobmsg_add_string(&bb, "action", action);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_uspd.trans_id);

	uspd_call(g_uspd.ubus_ctx, "operate", &bb, _sync_operate_cb, arg);

	blob_buf_free(&bb);

	return arg->fault;
}

static void _operate_async_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *results, *rtemp;
	int *instance = (int *) req->priv;
	char err_log[128] = "Error not specified by vendor";
	char *err_msg;
	int fault = USP_ERR_OK;
	kv_vector_t *output_args;
	size_t rem;

	output_args = USP_ARG_Create();
	if (!msg) {
		fault = USP_ERR_INTERNAL_ERROR;
		goto exit;
	}

	results = get_parameters(msg);
	if (results == NULL) {
		fault = USP_ERR_INTERNAL_ERROR;
		goto exit;
	}

	blobmsg_for_each_attr(rtemp, results, rem) {
		fault = detect_fault(rtemp);
		if (fault != USP_ERR_OK)
			break;

		struct blob_attr *params = get_parameters(rtemp);

		if (params == NULL)
			break;

		get_parameters_in_kv(params, output_args);
		break;
	}

exit:
	USP_LOG_Info("## Dump async out arg:");
	KV_VECTOR_Dump(output_args);

	err_msg = (fault != USP_ERR_OK) ? err_log : NULL;
	USP_SIGNAL_OperationComplete(*instance, fault, err_msg, output_args);
	USP_SAFE_FREE(instance);
}

static int _operate_async_handler(vendor_data_t *arg)
{
	struct blob_buf bb = {};
	int fault = USP_ERR_OK;
	char path[MAX_DM_PATH] = {'\0'}, *action;
	int *instance;

	instance = (int *) USP_MALLOC(sizeof(int));
	if (instance == NULL)
		return USP_ERR_COMMAND_FAILURE;

	// Exit if unable to signal that this operation is active
	fault = USP_SIGNAL_OperationStatus(arg->inst, "Active");
	if (fault != USP_ERR_OK) {
		USP_FREE(instance);
		return USP_ERR_COMMAND_FAILURE;
	}

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	// separate path and command
	action = TEXT_UTILS_SplitPath(arg->path, path, MAX_DM_PATH);

	blobmsg_add_string(&bb, "path", path);
	blobmsg_add_string(&bb, "action", action);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_uspd.trans_id);

	if (arg->kv != NULL) {
		void *table = blobmsg_open_table(&bb, "input");

		for (int i = 0; i < arg->kv->num_entries; ++i) {
			kv_pair_t *kv = &arg->kv->vector[i];

			blobmsg_add_string(&bb, kv->key, kv->value);
		}
		blobmsg_close_table(&bb, table);
	}

	*instance = arg->inst;
	fault = uspd_call_async("operate", &bb, _operate_async_cb, instance);
	if (fault != USP_ERR_OK) {
		USP_FREE(instance);
		arg->fault = fault;
	}

	blob_buf_free(&bb);
	return fault;
}

static void _tran_exec_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	bool status = false;
	struct blob_attr *f_attr[1] = {NULL};
	const struct blobmsg_policy fp[1] = {
		{ "status", BLOBMSG_TYPE_BOOL }
	};

	g_uspd.trans_id = 0;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_parse(fp, 1, f_attr, blobmsg_data(msg), blobmsg_len(msg));
	if (f_attr[0])
		status = blobmsg_get_u8(f_attr[0]);

	if (status == false) {
		USP_LOG_Error("Another transaction in progress");
		arg->fault = USP_ERR_INTERNAL_ERROR;
	}
}

static void _tran_start_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	bool status = false;
	int trans_id = 0;
	struct blob_attr *f_attr[2] = {NULL};
	const struct blobmsg_policy fp[2] = {
		{ "status", BLOBMSG_TYPE_BOOL },
		{ "transaction_id", BLOBMSG_TYPE_INT32 }
	};

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_parse(fp, 2, f_attr, blobmsg_data(msg), blobmsg_len(msg));

	if (f_attr[0])
		status = blobmsg_get_u8(f_attr[0]);

	if (f_attr[1])
		trans_id = blobmsg_get_u32(f_attr[1]);

	if (status == false) {
		USP_LOG_Info("Another transaction in progress");
		arg->fault = USP_ERR_INTERNAL_ERROR;
	} else {
		g_uspd.trans_id = trans_id;
	}
}

static int _tran_exec(vendor_data_t *arg, char *cmd)
{
	struct blob_buf bb = { };

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	// Invoke Ubus to get data from uspd
	if (strcmp(cmd, "transaction_start") == 0) {
		blobmsg_add_string(&bb, "app", "obuspa");
		blobmsg_add_u32(&bb, "max_timeout", g_uspd.ipc_timeout);
		uspd_call(g_uspd.ubus_ctx, cmd, &bb, _tran_start_cb, arg);
	} else {
		blobmsg_add_u32(&bb, "transaction_id", g_uspd.trans_id);
		uspd_call(g_uspd.ubus_ctx, cmd, &bb, _tran_exec_cb, arg);
	}

	blob_buf_free(&bb);
	return arg->fault;
}

static void _get_value_single_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	enum {
		P_PARAM,
		P_VALUE,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "parameter", BLOBMSG_TYPE_STRING },
		{ "value", BLOBMSG_TYPE_STRING }
	};

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	arg->fault = detect_fault(msg);
	if (arg->fault)
		return;

	params = get_parameters(msg);
	if (params == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL, NULL};
		char *path, *val;

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (!tb[P_PARAM] || !tb[P_VALUE])
			continue;

		path = blobmsg_get_string(tb[P_PARAM]);
		val = blobmsg_get_string(tb[P_VALUE]);

		if (strcmp(path, arg->path) == 0) {
			USP_STRNCPY(arg->p_value, val, arg->len);
			break;
		}
	}
}

static int _get_value_single(vendor_data_t *arg)
{
	struct blob_buf bb = { };

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "proto", "usp");

	// Invoke Ubus to get data from uspd
	uspd_call(g_uspd.ubus_ctx, "get", &bb, _get_value_single_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static void _get_value_group_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	int index, i;
	vendor_data_t *arg = req->priv;
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	kv_vector_t pv_set;
	enum {
		P_PARAM,
		P_VALUE,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "parameter", BLOBMSG_TYPE_STRING },
		{ "value", BLOBMSG_TYPE_STRING }
	};

	USP_ARG_Init(&pv_set);
	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	arg->fault = detect_fault(msg);
	if (arg->fault)
		return;

	params = get_parameters(msg);
	if (params == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	index = 0;
	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL, NULL};
		char *path, *val;

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (!tb[P_PARAM] || !tb[P_VALUE])
			continue;

		path = blobmsg_get_string(tb[P_PARAM]);
		val = blobmsg_get_string(tb[P_VALUE]);

		USP_ARG_Add(&pv_set, path, val);
	}

	for (i = 0, index = 0; i < arg->kv->num_entries; i++) {
		kv_pair_t *pair = &arg->kv->vector[i];

		index = KV_VECTOR_FindKey(&pv_set, pair->key, index);
		if (index != INVALID) {
			// Found a matching key, so replace its value
			USP_SAFE_FREE(pair->value);
			pair->value = USP_STRDUP(pv_set.vector[index].value);
			index++;
		} else {
			// If value not found meaning this instance does not exist
			// in lower layer, log error and return empty
			USP_SAFE_FREE(pair->value);
			pair->value = USP_STRDUP("");
			USP_LOG_Info("No value returned for [%s] from uspd", pair->key);
		}
	}

	if (arg->kv_out) {
		arg->kv_out->num_entries = pv_set.num_entries;
		arg->kv_out->vector = pv_set.vector;
		USP_ARG_Init(&pv_set);
	}

	USP_ARG_Destroy(&pv_set);
}

static int _get_value_group(vendor_data_t *arg)
{
	struct blob_buf bb = { };
	char *method, *p;
	void *array;
	int i;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "proto", "usp");
	if (strcmp(arg->path, ROOT_PATH) == 0) {
		method = "get";
		blobmsg_add_string(&bb, "path", arg->path);
	} else if (arg->kv->num_entries < MIN_NUM_TO_GROUP) {
		method = "getm_values";
		array = blobmsg_open_array(&bb, "paths");
		for (i = 0; i < arg->kv->num_entries; ++i) {
			p = arg->kv->vector[i].key;
			blobmsg_add_string(&bb, NULL, p);
		}
		blobmsg_close_array(&bb, array);
	} else {
		method = "get";
		blobmsg_add_string(&bb, "path", arg->path);
	}

	uspd_call(g_uspd.ubus_ctx, method, &bb, _get_value_group_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static int ubus_enque_cmd(vendor_data_t *arg)
{
	USP_LOG_Debug("## CMD (%s) path(%s) ##", stringify_cmd(arg->cmd), arg->path);
	if (IS_UBUS_Q_READY) {
		int bytes;

		pthread_mutex_lock(&g_uspd.ubus_lock);
		bytes = send(mq_tx_socket, &arg, sizeof(vendor_data_t *), 0);
		if (bytes == -1) {
			pthread_mutex_unlock(&g_uspd.ubus_lock);
			USP_LOG_Error("Failed to send err %s", strerror(errno));
			return USP_ERR_INTERNAL_ERROR;
		}

		if (bytes != sizeof(vendor_data_t *)) {
			pthread_mutex_unlock(&g_uspd.ubus_lock);
			USP_LOG_Error("[DMThread] send size miss-match (%d:%zd) to ubus thread", bytes, sizeof(vendor_data_t *));
			return USP_ERR_INTERNAL_ERROR;
		}
		pthread_cond_wait(&g_uspd.ubus_cond, &g_uspd.ubus_lock);
		pthread_mutex_unlock(&g_uspd.ubus_lock);

		return arg->fault;
	}

	switch (arg->cmd) {
	case CMD_GET:
		_get_value_single(arg);
		break;
	case CMD_GROUP_GET:
		_get_value_group(arg);
		break;
	case CMD_TRAN_START:
		_tran_exec(arg, "transaction_start");
		break;
	case CMD_TRAN_COMMIT:
		_tran_exec(arg, "transaction_commit");
		break;
	case CMD_TRAN_ABORT:
		_tran_exec(arg, "transaction_abort");
		break;
	default:
		USP_ERR_SetMessage("%s(%d): uspd thread not avilable", __FUNCTION__, __LINE__);
		arg->fault = USP_ERR_INTERNAL_ERROR;
		break;
	}

	return arg->fault;
}

static int uspd_operate_async(dm_req_t *req, kv_vector_t *input_args, int instance)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_OPERATE_ASYNC;
	arg.path = req->path;
	arg.kv = input_args;
	arg.inst = instance;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static int uspd_operate_sync(dm_req_t *req, char *command_key,
		      kv_vector_t *input_args, kv_vector_t *output_args)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_OPERATE_SYNC;
	arg.path = req->path;
	arg.kv = input_args;
	arg.kv_out = output_args;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static void _get_instances_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	struct blob_attr *cur, *params = NULL;
	size_t rem;
	enum {
		P_PARAM,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "parameter", BLOBMSG_TYPE_STRING }
	};

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	params = get_parameters(msg);
	if (params == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_PARAM]) {
			char *path = blobmsg_get_string(tb[P_PARAM]);

			STR_VECTOR_Add(arg->vec, path);
		}
	}
}

static int _get_instances(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "proto", "usp");

	uspd_call(g_uspd.ubus_ctx, "instances", &bb, _get_instances_cb, arg);

	blob_buf_free(&bb);

	return USP_ERR_OK;
}

static void register_uniq_param(struct blob_attr *blob_spath, struct blob_attr *uniq_keys)
{
	int num_keys = 0;
	char **keys, *spath;

	if (blob_spath == NULL || uniq_keys == NULL)
		return;

	spath = blobmsg_get_string(blob_spath);
	keys = blob_array_str_list(uniq_keys, &num_keys);
	if (num_keys != INVALID) {
		int i;
		int reg_key_num = 0;
		char **reg_keys;

		reg_keys = (char **) USP_MALLOC(num_keys * sizeof(char *));
		USP_ASSERT(reg_keys != NULL);

		// check if key registered as leaf
		for (i = 0; i < num_keys; i++) {
			char sparameter[MAX_DM_PATH] = {0};

			USP_SNPRINTF(sparameter, MAX_DM_PATH, "%s%s", spath, keys[i]);
			if (dm_node_present(sparameter) != NULL) {
				reg_keys[reg_key_num++] = keys[i];
			}
		}

		USP_REGISTER_Object_UniqueKey(spath, reg_keys, reg_key_num);
		USP_FREE(reg_keys);
	}

	USP_FREE(keys);
}

static void register_objects_uniq_params(struct blob_attr *blob_schema)
{
	struct blob_attr *cur;
	size_t rem;

	blobmsg_for_each_attr(cur, blob_schema, rem) {
		enum {
			P_PARAM,
			P_UNIQ,
			P_TYPE,
			__P_MAX
		};
		const struct blobmsg_policy p[__P_MAX] = {
			[P_PARAM] = { "parameter", BLOBMSG_TYPE_STRING },
			[P_UNIQ] = { "unique_keys", BLOBMSG_TYPE_ARRAY },
			[P_TYPE] = { "type", BLOBMSG_TYPE_STRING },
		};
		struct blob_attr *tb[__P_MAX] = {NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_PARAM] == NULL || tb[P_UNIQ] == NULL || tb[P_TYPE] == NULL) {
			continue;
		}

		if (get_dmt_type(blobmsg_get_string(tb[P_TYPE])) != DMT_OBJECT)
			continue;

		register_uniq_param(tb[P_PARAM], tb[P_UNIQ]);
	}
}

static int get_path_from_group_schema(char *group_name, char *schema)
{
	size_t i, len, j;

	j = 0;
	len = strlen(schema);
	group_name[0] = '\0';

	for (i = 0; i < len; i++) {
		if (schema[i] == '{') {
			group_name[j++] = '*';
			i += 2;
		} else {
			group_name[j++] = schema[i];
		}
	}

	// remove the last *
	if (group_name[j - 1] == '*')
		group_name[j - 1] = '\0';

	return 0;
}

static bool is_cached_data(kv_vector_t *params)
{
	kv_vector_t *kv;
	int i;
	bool all_found = false;

	kv = &g_uspd.cached_kv;
	if (kv->num_entries == 0)
		return false;

	for (i = 0; i < params->num_entries; i++) {
		kv_pair_t *pair = &params->vector[i];

		g_uspd.cached_index = KV_VECTOR_FindKey(kv, pair->key, g_uspd.cached_index);
		if (g_uspd.cached_index != INVALID) {
			USP_SAFE_FREE(pair->value);
			pair->value = USP_STRDUP(kv->vector[g_uspd.cached_index].value);
			g_uspd.cached_index++;
			all_found = true;
		} else {
			all_found = false;
			break;
		}
	}

	return all_found;
}

// DM Caching algorithm
// ret =>
//    false: data not valid, get fresh data
//    true: data present in cached and updated in ckv
static bool check_data_inventory(int gid, kv_vector_t *ckv)
{
	unsigned t;

	if (g_uspd.cached_kv.num_entries == 0)
		return false;

	t = tu_uptime_secs() - g_uspd.last_cached_time;

	// time check
	if (t >= g_uspd.cached_interval) {
		// USP_LOG_Debug("Dropping cached data as its old by %u", t);
		KV_VECTOR_Destroy(&g_uspd.cached_kv);
		return false;
	}

	// Access non group params
	if ((gid == INVALID) && (g_uspd.last_gid == INVALID) && is_cached_data(ckv))
		return true;

	if (gid != INVALID && g_uspd.last_gid == gid) {
		// USP_LOG_Debug("Dropping cached data as its same group %d", gid);
		KV_VECTOR_Destroy(&g_uspd.cached_kv);
		return false;
	}

	// out of sequence access check
	if (g_uspd.last_gid > gid) {
		// USP_LOG_Debug("Dropping cached data out-of-sequence last:current[%d:%d]", g_uspd.last_gid, gid);
		KV_VECTOR_Destroy(&g_uspd.cached_kv);
		return false;
	}

	// Cache is not present
	if (is_cached_data(ckv)) {
		return true;
	}

	return false;
}

static int group_get(int group_id, kv_vector_t *params)
{
	char group_path[MAX_DM_PATH] = {0};
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));

	// If all parameters not found in cached data, clean cached data and get it from uspd
	if (check_data_inventory(group_id, params) == false) {
		get_path_from_group_schema(group_path, g_uspd.group_vec.vector[group_id]);
		arg.cmd = CMD_GROUP_GET;
		arg.path = group_path;
		arg.kv = params;
		ubus_enque_cmd(&arg);
	}

	g_uspd.last_gid = group_id;

	return arg.fault;
}

static int uspd_set_value(dm_req_t *req, char *buf)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_SET;
	arg.path = req->path;
	arg.p_value = buf;
	arg.len = strlen(buf);

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static int group_set(int group_id, kv_vector_t *params, unsigned *param_types, int *failure_index)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));

	*failure_index = INVALID;
	arg.cmd = CMD_GROUP_SET;
	arg.kv = params;
	arg.p_failure_index = failure_index;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static int group_add(int group_id, char *path, int *instance)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_ADD;
	arg.path = path;
	arg.p_instance = instance;

	ubus_enque_cmd(&arg);

	if (arg.fault != 0)
		return USP_ERR_CREATION_FAILURE;

	return 0;
}

static int group_del(int group_id, char *path)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_DEL;
	arg.path = path;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static unsigned int get_cached_time(char *path)
{
	int i;

	for (i = 0; i < g_uspd.cached_exclude_vec.num_entries; i++) {
		size_t len = strlen(g_uspd.cached_exclude_vec.vector[i]);

		if (strncmp(path, g_uspd.cached_exclude_vec.vector[i], len) == 0) {
			return 0;
		}
	}

	return g_uspd.cached_interval;
}

static int refresh_instances(int group_id, char *path, int *expiry_period)
{
	int i;
	vendor_data_t arg;
	str_vector_t inst_vec;

	memset(&arg, 0, sizeof(vendor_data_t));

	STR_VECTOR_Init(&inst_vec);
	arg.cmd = CMD_INSTANCES;
	arg.path = path;
	arg.vec = &inst_vec;

	ubus_enque_cmd(&arg);

	for (i = 0; i < inst_vec.num_entries; i++) {
		USP_DM_RefreshInstance(inst_vec.vector[i]);
	}

	*expiry_period = get_cached_time(path);
	STR_VECTOR_Destroy(&inst_vec);

	// Drop cached data
	KV_VECTOR_Destroy(&g_uspd.cached_kv);

	return USP_ERR_OK;
}

static void update_instance_refresh_vector(char *path, str_vector_t *refresh_vec)
{
	int i;
	bool ref_found;

	ref_found = false;
	for (i = 0; i < refresh_vec->num_entries ; i++) {
		size_t len = strlen(refresh_vec->vector[i]);

		if (strncmp(path, refresh_vec->vector[i], len) == 0) {
			ref_found = true;
			break;
		}
	}

	if (ref_found == false)
		STR_VECTOR_Add(refresh_vec, path);
}

static bool get_bool_from_str(char *bool_str)
{
	bool ret = false;

	if (bool_str == NULL)
		return false;

	if (TEXT_UTILS_StringToBool(bool_str, &ret) != USP_ERR_OK) {
		USP_LOG_Info("Decode perm of some param failed");
		return false;
	}

	return ret;
}
static int create_dm_group(char *spath, char *dm_type, char *dm_write, str_vector_t *refresh_vec)
{
	bool ret = false;
	int index, fault;
	dmt_type_t bbf_dmt;
	char group_name[MAX_DM_PATH];
	bool is_writable;

	if (dm_write[0] == '\0') {
		return false;
	}

	if (CountPathSeparator(spath) < MAX_GROUP_SEP) {
		return false;
	}

	TEXT_UTILS_SplitPathAtSeparator(spath, group_name, MAX_DM_PATH, 2);
	index = STR_VECTOR_Find(&g_uspd.group_vec, group_name);
	is_writable = get_bool_from_str(dm_write);

	// Add group if does not exists
	if (index == INVALID) {
		char tmp[MAX_DM_PATH];
		int group_id = g_uspd.group_vec.num_entries;

		USP_SNPRINTF(tmp, MAX_DM_PATH, "%s{i}.", group_name);
		if (strcmp(spath, tmp) == 0) {
			fault = USP_REGISTER_GroupedObject(group_id, spath, is_writable);
			update_instance_refresh_vector(tmp, refresh_vec);
			ret = true;
		} else {
			fault = USP_REGISTER_GroupedObject(group_id, group_name, 0);
			ret = false;
		}

		if (fault == USP_ERR_OK) {
			// USP_LOG_Debug("# ADD GROUP [%d => %s => %s] #", group_id, group_name, spath);
			USP_REGISTER_GroupVendorHooks(group_id, group_get, group_set, group_add, group_del);
			STR_VECTOR_Add(&g_uspd.group_vec, group_name);
			index = group_id;
		}
	}

	bbf_dmt = get_dmt_type(dm_type);
	if (bbf_dmt != DMT_OBJECT) {
		return fault;
	}

	if (index != INVALID && ret == false) {
		fault = USP_REGISTER_GroupedObject(index, spath, is_writable);
		if (fault == USP_ERR_OK) {
			update_instance_refresh_vector(spath, refresh_vec);
			// USP_LOG_Debug("## Enroll [%s <= %s] OBJ", g_uspd.group_vec.vector[index], spath);
		}
	}

	return fault;
}

static int get_schema_params_from_blob(struct blob_attr *param_blob, char *param, char *writable, char *type)
{
	enum {
		P_PARAM,
		P_WRITEABLE,
		P_TYPE,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		[P_PARAM] = { "parameter", BLOBMSG_TYPE_STRING },
		[P_WRITEABLE] = { "writable", BLOBMSG_TYPE_STRING },
		[P_TYPE] = { "type", BLOBMSG_TYPE_STRING },
	};
	struct blob_attr *tb[__P_MAX] = {NULL};

	if (param_blob == NULL) {
		USP_LOG_Error("Empty parameter blob");
		return INVALID;
	}

	if (blobmsg_type(param_blob) != BLOBMSG_TYPE_TABLE) {
		USP_LOG_Error("Parameter is not an table");
		return INVALID;
	}

	blobmsg_parse(p, __P_MAX, tb, blobmsg_data(param_blob), blobmsg_len(param_blob));
	if (!tb[P_PARAM] || !tb[P_TYPE]) {
		USP_LOG_Error("mandatory information missing in param blob");
		return INVALID;
	}

	USP_STRNCPY(param, blobmsg_get_string(tb[P_PARAM]), MAX_DM_PATH);
	USP_STRNCPY(type, blobmsg_get_string(tb[P_TYPE]), MAX_DM_PATH);
	if (tb[P_WRITEABLE] != NULL) {
		USP_STRNCPY(writable, blobmsg_get_string(tb[P_WRITEABLE]), MAX_DM_PATH);
	}

	return USP_ERR_OK;
}

static dmt_type_t get_dmt_type(char *type)
{
	int i;

	if (type == NULL) {
		USP_LOG_Error("Type info missing the dm parameter");
		return DM_STRING;
	}

	for (i = 0; i < ARRAY_SIZE(DMT_STR_MAP); i++) {
		if (strcmp(type, DMT_STR_MAP[i]) == 0)
			return i;
	}

	USP_LOG_Warning("Type info not matching, defaults to string");
	return DM_STRING;
}

static bool skip_core_dm(char *spath)
{
	if ((strncmp(spath, "Device.BulkData.", 16) == 0) ||
	    (strncmp(spath, "Device.Security.", 16) == 0))
		return true;

	return false;
}

static int create_dm_groups(struct blob_attr *param_blob, str_vector_t *refresh_vec)
{
	struct blob_attr *cur;
	size_t rem;

	blobmsg_for_each_attr(cur, param_blob, rem) {
		char dm_path[MAX_DM_PATH] = {0};
		char dm_type[MAX_DM_PATH] = {0};
		char dm_write[MAX_DM_PATH] = {0};
		int ret;

		ret = get_schema_params_from_blob(cur, dm_path, dm_write, dm_type);
		if (ret == INVALID) {
			USP_LOG_Error("Empty parameter to handle");
			continue;
		}

		if (skip_core_dm(dm_path)) {
			continue;
		}

		create_dm_group(dm_path, dm_type, dm_write, refresh_vec);
	}
	return USP_ERR_OK;
}

static int convert_dmt_to_dmtype(dmt_type_t dmt_type)
{
	int dm_type = INVALID;

	switch(dmt_type) {
		case DMT_STRING:
			dm_type = DM_STRING;
			break;
		case DMT_UNINT:
			dm_type = DM_UINT;
			break;
		case DMT_INT:
			dm_type = DM_INT;
			break;
		case DMT_UNLONG:
			dm_type = DM_ULONG;
			break;
		case DMT_LONG:
			dm_type = DM_ULONG;
			break;
		case DMT_BOOL:
			dm_type = DM_BOOL;
			break;
		case DMT_TIME:
			dm_type = DM_DATETIME;
			break;
		case DMT_HEXBIN:
			dm_type = DM_STRING;
			break;
		case DMT_BASE64:
			dm_type = DM_STRING;
			break;
		default:
			dm_type = DM_STRING;
	}

	return dm_type;
}

static int initiate_data_caching(kv_vector_t *get_kv)
{
	vendor_data_t arg;
	uint32_t start, stop;

	memset(&arg, 0, sizeof(vendor_data_t));

	arg.kv_out = &g_uspd.cached_kv;
	arg.cmd = CMD_GROUP_GET;
	arg.path = ROOT_PATH;
	arg.kv = get_kv;

	start = tu_uptime_msecs();
	ubus_enque_cmd(&arg);
	stop = tu_uptime_msecs();

	g_uspd.ipc_timeout = DEFAULT_IPC_TIMEOUT + (stop - start);
	g_uspd.last_cached_time = tu_uptime_secs();

	USP_LOG_Info("Updated IPC(Ubus) timeout %d", g_uspd.ipc_timeout);

	return arg.fault;
}

int uspd_get_value(dm_req_t *req, char *buf, int len)
{
	kv_vector_t kv_param;
	int fault = USP_ERR_OK;
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));

	USP_ARG_Init(&kv_param);
	USP_ARG_Add(&kv_param, req->path, NULL);

	// Get of complete tree is driven by DM tread, and hooks called to fill
	// specific sub-tree. This is hack to optimize overall get time
	// Check if Caching is required
	if (check_data_inventory(INVALID, &kv_param) == false) {
		fault = initiate_data_caching(&kv_param);
	}

	if (kv_param.num_entries && fault == USP_ERR_OK) {
		USP_STRNCPY(buf, kv_param.vector[0].value, len)
	}

	USP_ARG_Destroy(&kv_param);
	g_uspd.last_gid = INVALID;

	return fault;
}

bool register_num_entries(kv_vector_t *numentry_kv, str_vector_t *obj_list)
{
       int i, j;
       char param[MAX_DM_PATH];
       int fault = USP_ERR_OK;

       // check if the dm entry is a NumberOfEntries parameter
       for (i = 0; i < numentry_kv->num_entries; i++) {
               char *spath = numentry_kv->vector[i].key;
               char *obj = numentry_kv->vector[i].value;

               // prepare dm object with which it should associate
               USP_SNPRINTF(param, MAX_DM_PATH, "%s.{i}.", obj);

               dm_node_t *node = dm_node_present(param);
               if (node == NULL)
               {
		       for (j = 0; j < obj_list->num_entries; j++) {
			       size_t len = strlen(obj);

			       if (strncmp(obj_list->vector[j], obj, len) == 0) {
				       fault = USP_REGISTER_Param_NumEntries(spath, obj_list->vector[j]);
				       break;
			       }
		       }
		       continue;
               }

               // Exit if this is not an object
               if (IsObject(node)) {
		       fault = USP_REGISTER_Param_NumEntries(spath, param);
               } else {
		       USP_LOG_Info("%s Object %s not registered, skip uniq key reg", __FUNCTION__, param);
	       }
       }

       return fault;
}


bool handle_num_entries_reg(char *spath, kv_vector_t *numentry_kv)
{
	char *p;
	int len, num_len;
	char obj_path[MAX_DM_PATH] = {0};

	// check if the dm entry is a NumberOfEntries parameter
	len = strlen(spath);
	num_len = strlen(NUMENTRY);
	if (len < num_len || len > MAX_DM_PATH)
		return false;

	len = len - num_len;
	USP_STRNCPY(obj_path, spath, MAX_DM_PATH - 1);
	obj_path[len] = '.';
	obj_path[len + 1] = '\0';

	p = spath + len;
	if (strcmp(p, NUMENTRY) == 0) {
		KV_VECTOR_Add(numentry_kv, spath, obj_path);
		return true;
	}

	return false;
}

void uspd_register_leaf(char *spath, dmt_type_t bbf_dmt, char *dm_write)
{
	int group, i;
	int type;
	bool is_writable;

	type = convert_dmt_to_dmtype(bbf_dmt);
	// most likely the leaf is of last added group, so checking it in
	// reverse search
	group = INVALID;
	for (i = g_uspd.group_vec.num_entries - 1; i >= 0; i--) {
		char *p = g_uspd.group_vec.vector[i];
		size_t len = strlen(p);

		if (strncmp(spath, p, len) == 0) {
			group = i;
			break;
		}
	}

	is_writable = get_bool_from_str(dm_write);
	if (group == INVALID) {
		// USP_LOG_Debug("LEAF without group (%s)", spath);
		if (is_writable)
			USP_REGISTER_VendorParam_ReadWrite(spath, uspd_get_value, uspd_set_value, NULL, 0);
		else
			USP_REGISTER_VendorParam_ReadOnly(spath, uspd_get_value, 0);
	} else {
		// USP_LOG_Debug("# LEAF (%d => %s)", group, spath);
		if (is_writable)
			USP_REGISTER_GroupedVendorParam_ReadWrite(group, spath, type);
		else
			USP_REGISTER_GroupedVendorParam_ReadOnly(group, spath, type);
	}

}

static bool handle_prereg_dm(kv_vector_t *cc_param, char *spath, dmt_type_t bbf_type, kv_vector_t *num_entries_kv)
{
	char *val;

	if (skip_core_dm(spath))
		return true;

	if (handle_num_entries_reg(spath, num_entries_kv))
		return true;

	val = KV_VECTOR_Get(cc_param, spath, NULL, 0);
	if (val) {
		int type;

		type = convert_dmt_to_dmtype(bbf_type);
		USP_REGISTER_Param_Constant(spath, val, type);
		return true;
	}

	return false;
}

static int cache_const_dm(kv_vector_t *cached_kv)
{
	int i;
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));

	for (i = 0; i < ARRAY_SIZE(CONST_DM); i++) {
		USP_ARG_Add(cached_kv, CONST_DM[i], "");
	}

	arg.kv = cached_kv;
	arg.cmd = CMD_GROUP_GET;
	arg.path = "Device.DeviceInfo.";

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static void register_schema_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *schema_blob = NULL;
	struct blob_attr *params;
	struct blob_attr *cur;
	vendor_data_t *arg = req->priv;
	str_vector_t ref_param_vec;
	kv_vector_t num_entries_kv;
	size_t rem = 0, i;
	int ret;
	kv_vector_t cached_kv;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	schema_blob = blob_memdup(msg);
	if (schema_blob == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	STR_VECTOR_Init(&ref_param_vec);
	KV_VECTOR_Init(&num_entries_kv);
	KV_VECTOR_Init(&cached_kv);

	params = get_parameters(schema_blob);
	if (params == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		goto end;
	}

	ret = create_dm_groups(params, &ref_param_vec);
	if (ret != USP_ERR_OK) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		goto end;
	}

	// Get const datamodel value
	ret = cache_const_dm(&cached_kv);
	if (ret != USP_ERR_OK) {
		USP_LOG_Error("Error getting mandatory data, ret %d", ret);
		arg->fault = ret;
		goto end;
	}

	blobmsg_for_each_attr(cur, params, rem) {
		char dm_spath[MAX_DM_PATH] = {0};
		char dm_stype[MAX_DM_PATH] = {0};
		char dm_write[MAX_DM_PATH] = {0};

		dmt_type_t bbf_dmt;

		ret = get_schema_params_from_blob(cur, dm_spath, dm_write, dm_stype);
		if (ret == INVALID) {
			USP_LOG_Error("Empty parameter blob");
			continue;
		}

		bbf_dmt = get_dmt_type(dm_stype);
		switch(bbf_dmt) {
			case DMT_COMMAND: // Register commands
			case DMT_EVENT: // Register events
			{
				register_operates_events(cur, bbf_dmt);
				break;
			}
			case DMT_OBJECT: // Skip objects
			{
				break;
			}
			default: // Register leaf
			{
				if (handle_prereg_dm(&cached_kv, dm_spath, bbf_dmt, &num_entries_kv) == false) {
					uspd_register_leaf(dm_spath, bbf_dmt, dm_write);
				}
			}

		}


	}

	// register uniq parameters
	register_objects_uniq_params(params);

	// register refresh instances
	for (i = 0; i < ref_param_vec.num_entries; ++i) {
		// USP_LOG_Debug("REF Inst(%s)", ref_param_vec.vector[i]);
		USP_REGISTER_Object_RefreshInstances(ref_param_vec.vector[i], refresh_instances);
	}

	register_num_entries(&num_entries_kv, &ref_param_vec);

end:
	KV_VECTOR_Destroy(&cached_kv);
	KV_VECTOR_Destroy(&num_entries_kv);
	STR_VECTOR_Destroy(&ref_param_vec);
	USP_FREE(schema_blob);
}

static char **blob_array_str_list(struct blob_attr *attr, int *list_nr)
{
	struct blob_attr *cur;
	int len, i;
	char **out;
	size_t rem;

	*list_nr = INVALID;

	len = blobmsg_check_array(attr, BLOBMSG_TYPE_STRING);
	out = (char **) USP_MALLOC(len * sizeof(char *));
	USP_ASSERT(out != NULL);

	i = 0;
	blobmsg_for_each_attr(cur, attr, rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_STRING)
			continue;

		out[i++] = blobmsg_get_string(cur);
	}

	*list_nr = i;

	return out;
}

static int register_event_dm(struct blob_attr *param, struct blob_attr *in_attr)
{
	char *epath;
	int in_nr = 0, fault;
	char **in = NULL;

	if (param == NULL || in_attr == NULL)
		return INVALID;

	epath = blobmsg_get_string(param);
	// USP_LOG_Debug("Registering event [%s]", epath);

	fault = USP_REGISTER_Event(epath);
	if (fault != USP_ERR_OK) {
		return fault;
	}

	in = blob_array_str_list(in_attr, &in_nr);
	if (in_nr != INVALID) {
		fault = USP_REGISTER_EventArguments(epath, in, in_nr);
	}

	USP_FREE(in);
	return fault;
}

static int register_operate_dm(struct blob_attr *param, struct blob_attr *type,
		struct blob_attr *in_attr, struct blob_attr *out_attr)
{
	char *path;
	char *op_type;
	int fault = USP_ERR_OK;
	char **in = NULL, **out = NULL;
	int in_nr = 0, out_nr = 0;

	if (param == NULL || type == NULL) {
		USP_LOG_Error("Type info not present");
		return USP_ERR_INTERNAL_ERROR;
	}

	path = blobmsg_get_string(param);
	op_type = blobmsg_get_string(type);
	// USP_LOG_Debug("Registering operate [%s => %s]", path, op_type);

	if (dm_node_present(path) != NULL) {
		// USP_LOG_Debug("Operation (%s) already registered", path);
		return USP_ERR_OK;
	}
	if (strncmp(op_type, "sync", 4) == 0) {
		fault = USP_REGISTER_SyncOperation(path, uspd_operate_sync);
	} else {
		fault = USP_REGISTER_AsyncOperation(path, uspd_operate_async, NULL);
	}

	if (fault != USP_ERR_OK) {
		return fault;
	}

	if (in_attr != NULL)
		in = blob_array_str_list(in_attr, &in_nr);

	if (out_attr != NULL)
		out = blob_array_str_list(out_attr, &out_nr);

	fault = USP_REGISTER_OperationArguments(path, in, in_nr, out, out_nr);

	USP_FREE(in);
	USP_FREE(out);

	return fault;
}

static void register_operates_events(struct blob_attr *param_blob, dmt_type_t bbf_type)
{
	enum {
		P_PARAM,
		P_INPUT,
		P_OUTPUT,
		P_OP_TYPE,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		[P_PARAM] = {"parameter", BLOBMSG_TYPE_STRING},
		[P_INPUT] = {"in", BLOBMSG_TYPE_ARRAY},
		[P_OUTPUT] = {"out", BLOBMSG_TYPE_ARRAY},
		[P_OP_TYPE] = {"cmd_type", BLOBMSG_TYPE_STRING},
	};
	struct blob_attr *tb[__P_MAX];

	if (param_blob == NULL)
		return;

	blobmsg_parse(p, __P_MAX, tb, blobmsg_data(param_blob), blobmsg_len(param_blob));
	if (!tb[P_PARAM]) {
		return;
	}

	if (bbf_type == DMT_EVENT) {
		register_event_dm(tb[P_PARAM], tb[P_INPUT]);
	} else if (bbf_type == DMT_COMMAND) {
		register_operate_dm(tb[P_PARAM], tb[P_OP_TYPE], tb[P_INPUT], tb[P_OUTPUT]);
	} else {
		USP_LOG_Error("Invalid type for registration");
	}
}

static void read_cmd_callback(struct uloop_fd *u, unsigned int events)
{
	vendor_data_t *arg = NULL;
	int bytes_read;

	pthread_mutex_lock(&g_uspd.ubus_lock);
	bytes_read = recv(mq_rx_socket, &arg, sizeof(vendor_data_t *), 0);
	if (bytes_read != sizeof(vendor_data_t *)) {
		USP_LOG_Error("[UBUS] size mismatch %d", bytes_read);
		pthread_cond_signal(&g_uspd.ubus_cond);
		pthread_mutex_unlock(&g_uspd.ubus_lock);
		return;
	}

	switch (arg->cmd) {
	case CMD_GET:
		_get_value_single(arg);
		break;
	case CMD_GROUP_GET:
		_get_value_group(arg);
		break;
	case CMD_INSTANCES:
		_get_instances(arg);
		break;
	case CMD_ADD:
		_add_object(arg);
		break;
	case CMD_SET:
		_set_value_single(arg);
		break;
	case CMD_GROUP_SET:
		_set_value_group(arg);
		break;
	case CMD_DEL:
		_del_object(arg);
		break;
	case CMD_OPERATE_SYNC:
		_operate_sync_handler(arg);
		break;
	case CMD_OPERATE_ASYNC:
		_operate_async_handler(arg);
		break;
	case CMD_TRAN_START:
		_tran_exec(arg, "transaction_start");
		break;
	case CMD_TRAN_COMMIT:
		_tran_exec(arg, "transaction_commit");
		break;
	case CMD_TRAN_ABORT:
		_tran_exec(arg, "transaction_abort");
		break;
	case CMD_NULL:
		USP_LOG_Info("Command not supported as of now");
		break;
	}

	pthread_cond_signal(&g_uspd.ubus_cond);
	pthread_mutex_unlock(&g_uspd.ubus_lock);
}

static kv_vector_t *blob_table_to_kv(struct blob_attr *table)
{
	kv_vector_t *params = NULL;
	struct blob_attr *cur;
	size_t rem;

	params = USP_ARG_Create();

	if (!table)
		goto exit;
	if (blob_id(table) != BLOBMSG_TYPE_TABLE)
		goto exit;

	rem = blobmsg_data_len(table);
	__blob_for_each_attr(cur, blobmsg_data(table), rem) {
		struct blobmsg_hdr *hdr;
		char value[MAX_DM_VALUE_LEN] = {0};

		hdr = blob_data(cur);
		switch (blob_id(cur)) {
			case BLOBMSG_TYPE_STRING:
				snprintf(value, MAX_DM_VALUE_LEN, "%s", blobmsg_get_string(cur));
				break;
			case BLOBMSG_TYPE_INT8:
				snprintf(value, MAX_DM_VALUE_LEN, "%d", blobmsg_get_u8(cur));
				break;
			case BLOBMSG_TYPE_INT16:
				snprintf(value, MAX_DM_VALUE_LEN, "%d", blobmsg_get_u16(cur));
				break;
			case BLOBMSG_TYPE_INT32:
				snprintf(value, MAX_DM_VALUE_LEN, "%u", blobmsg_get_u32(cur));
				break;
			case BLOBMSG_TYPE_INT64:
			case BLOBMSG_TYPE_DOUBLE:
				snprintf(value, MAX_DM_VALUE_LEN, "%"PRIu64"", blobmsg_get_u64(cur));
				break;
			default:
				USP_LOG_Info("Unhandled set request type|%x|", blob_id(cur));
		}
		USP_ARG_Add(params, (char *)hdr->name, value);
	}
exit:
	return params;
}

static void event_handler(struct ubus_context *ctx, struct ubus_event_handler *ev, const char *type, struct blob_attr *msg)
{
	char *ename;
	kv_vector_t *eparam;
	dm_node_t *node;
	enum {
		P_NAME,
		P_INPUT,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "name", BLOBMSG_TYPE_STRING },
		{ "input", BLOBMSG_TYPE_TABLE },
	};
	struct blob_attr *tb[__P_MAX] = {NULL};

	if (blobmsg_parse(p, __P_MAX, tb, blobmsg_data(msg), blobmsg_len(msg)) != UBUS_STATUS_OK) {
		USP_LOG_Info("Failed to parse ubus data");
		return;
	}

	if (!tb[P_NAME]) {
		USP_LOG_Info("Event name not present");
	}
	ename = blobmsg_get_string(tb[P_NAME]);
	node = DM_PRIV_GetNodeFromPath(ename, NULL, NULL);
	if (node == NULL) {
		USP_LOG_Info("Event/Param (%s) not registered", ename);
		return;
	}

	if (node->type != kDMNodeType_Event) {
		USP_LOG_Info("Parameter (%s) not an event", ename);
		return;
	}

	eparam = blob_table_to_kv(tb[P_INPUT]);
	// Dump parameters for cli validation
	KV_VECTOR_Dump(eparam);
	USP_SIGNAL_DataModelEvent(ename, eparam);
}

static void event_handler_add_del(struct ubus_context *ctx, struct ubus_event_handler *ev, const char *type, struct blob_attr *msg)
{
	bool is_add = false;
	struct blob_attr *cur;
	int rem;
	enum {
		P_PARAMS,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "instances", BLOBMSG_TYPE_ARRAY },
	};
	struct blob_attr *tb[__P_MAX] = {NULL};

	if (blobmsg_parse(p, __P_MAX, tb, blobmsg_data(msg), blobmsg_len(msg)) != UBUS_STATUS_OK) {
		USP_LOG_Info("Failed to parse ubus data");
		return;
	}

	if (!tb[P_PARAMS]) {
		USP_LOG_Info("Event %s has no instances to update", type);
		return;
	}

	if (strcmp(type, USPD_ADD_OBJ_EV_NAME) == 0)
		is_add = true;

	blobmsg_for_each_attr(cur, tb[P_PARAMS], rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_STRING)
			continue;

		if (is_add)
			USP_SIGNAL_ObjectAdded(blobmsg_get_string(cur));
		else
			USP_SIGNAL_ObjectDeleted(blobmsg_get_string(cur));
	}
}

static void *start_ubus_thread(void *arg)
{
	struct uloop_fd cmd_fd;

	uloop_init();
	ubus_add_uloop(g_uspd.ubus_ctx);

	memset(&cmd_fd, 0, sizeof(struct uloop_fd));
	cmd_fd.cb = read_cmd_callback;
	cmd_fd.fd = mq_rx_socket;
	uloop_fd_add(&cmd_fd, ULOOP_READ | ULOOP_EDGE_TRIGGER);

	// register for ubus events
	g_uspd.ubus_notify_event.cb = event_handler;
	ubus_register_event_handler(g_uspd.ubus_ctx, &g_uspd.ubus_notify_event, USPD_EVENT);

	g_uspd.ubus_add_event.cb = event_handler_add_del;
	ubus_register_event_handler(g_uspd.ubus_ctx, &g_uspd.ubus_add_event, USPD_ADD_OBJ_EV_NAME);
	g_uspd.ubus_del_event.cb = event_handler_add_del;
	ubus_register_event_handler(g_uspd.ubus_ctx, &g_uspd.ubus_del_event, USPD_DEL_OBJ_EV_NAME);

	uloop_run();

	uloop_done();

	close(mq_tx_socket);
	close(mq_rx_socket);

	return NULL;
}

static int vendor_exec_init(void)
{
    int err;

    // Exit if unable to initialize the unix domain socket pair used to implement a message queue
    err = socketpair(AF_UNIX, SOCK_DGRAM, 0, g_uspd.mq_sockets);
    if (err != 0)
    {
        USP_ERR_ERRNO("socketpair", errno);
        return USP_ERR_INTERNAL_ERROR;
    }

    return USP_ERR_OK;
}

static int uspd_register_schema()
{
	struct blob_buf bb = {};
	int fault = USP_ERR_OK;
	vendor_data_t arg;

	// register mandatory parameters
	memset(&arg, 0, sizeof(vendor_data_t));

	// Invoke Ubus to get data from uspd
	blob_buf_init(&bb, 0);
	blobmsg_add_string(&bb, "path", ROOT_PATH);
	blobmsg_add_u8(&bb, "next-level", true);
	blobmsg_add_u32(&bb, "schema_type", 0);
	fault = uspd_call(g_uspd.ubus_ctx, "get_supported_dm", &bb, register_schema_cb, &arg);
	blob_buf_free(&bb);

	return fault;
}

int factory_reset(void)
{
	int res = USP_ERR_OK;
	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", ROOT_PATH);
	blobmsg_add_string(&bb, "action", "FactoryReset()");

	res = uspd_call_inline_nocb("operate", &bb);
	blob_buf_free(&bb);

	return res;
}

int reboot(void)
{
	int res = USP_ERR_OK;
	struct blob_buf bb;


	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", ROOT_PATH);
	blobmsg_add_string(&bb, "action", "Reboot()");

	res = uspd_call_inline_nocb("operate", &bb);
	blob_buf_free(&bb);

	return res;
}

static int uspd_tran_start()
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_TRAN_START;

	if (g_uspd.op_state == USPD_INIT) {
		g_uspd.op_state = USPD_PRE_START;
		return USP_ERR_OK;
	}

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static int uspd_tran_commit()
{
	vendor_data_t arg;

	if (g_uspd.op_state == USPD_PRE_START) {
		g_uspd.op_state = USPD_START;
		return USP_ERR_OK;
	}

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_TRAN_COMMIT;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static int uspd_tran_abort()
{
	vendor_data_t arg;

	if (g_uspd.op_state == USPD_PRE_START) {
		g_uspd.op_state = USPD_START;
		return USP_ERR_OK;
	}
	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_TRAN_ABORT;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static unsigned short get_perm_mask(struct blob_attr *perm_arr)
{
	unsigned short perm_mask = PERMIT_NONE;
	char *perm_string;
	struct blob_attr *cur;
	int rem;

	blobmsg_for_each_attr(cur, perm_arr, rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_STRING) {
			USP_LOG_Debug("Wrong json format permission");
			continue;
		}

		perm_string = blobmsg_get_string(cur);

		if (strcmp(perm_string,"PERMIT_ALL") == 0) {
			perm_mask |= PERMIT_ALL;
		} else if (strcmp(perm_string,"PERMIT_GET") == 0) {
			perm_mask |= PERMIT_GET;
		} else if (strcmp(perm_string,"PERMIT_SET") == 0) {
			perm_mask |= PERMIT_SET;
		} else if (strcmp(perm_string,"PERMIT_ADD") == 0) {
			perm_mask |= PERMIT_ADD;
		} else if (strcmp(perm_string,"PERMIT_DEL") == 0) {
			perm_mask |= PERMIT_DEL;
		} else if (strcmp(perm_string,"PERMIT_OPER") == 0) {
			perm_mask |= PERMIT_OPER;
		} else if (strcmp(perm_string,"PERMIT_SUBS_VAL_CHANGE") == 0) {
			perm_mask |= PERMIT_SUBS_VAL_CHANGE;
		} else if (strcmp(perm_string,"PERMIT_SUBS_OBJ_ADD") == 0) {
			perm_mask |= PERMIT_SUBS_OBJ_ADD;
		} else if (strcmp(perm_string,"PERMIT_SUBS_OBJ_DEL") == 0) {
			perm_mask |= PERMIT_SUBS_OBJ_DEL;
		} else if (strcmp(perm_string,"PERMIT_SUBS_EVT_OPER_COMP") == 0) {
			perm_mask |= PERMIT_SUBS_EVT_OPER_COMP;
		} else if (strcmp(perm_string,"PERMIT_GET_INST") == 0) {
			perm_mask |= PERMIT_GET_INST;
		} else if (strcmp(perm_string,"PERMIT_OBJ_INFO") == 0) {
			perm_mask |= PERMIT_OBJ_INFO;
		} else if (strcmp(perm_string,"PERMIT_CMD_INFO") == 0) {
			perm_mask |= PERMIT_CMD_INFO;
		} else {
			perm_mask = PERMIT_NONE;
		}
	}

	return perm_mask;
}

static void create_rule_perm(int rule_index, struct blob_attr *perm)
{
	int err;
	unsigned short perm_bitmask = 0;
	enum {
		ATTR_OBJ,
		ATTR_PERM,
		__MAX
	};
	struct blob_attr *f_attr[__MAX] = {NULL};
	const struct blobmsg_policy fp[__MAX] = {
		{"object", BLOBMSG_TYPE_STRING},
		{"perm", BLOBMSG_TYPE_ARRAY}
	};

	if (blobmsg_parse(fp, __MAX, f_attr, blobmsg_data(perm), blobmsg_len(perm)))
		return;

	if (!f_attr[ATTR_OBJ] || !f_attr[ATTR_PERM])
		return;

	perm_bitmask = get_perm_mask(f_attr[ATTR_PERM]);
	err = USP_DM_AddControllerTrustPermission(rule_index, blobmsg_get_string(f_attr[ATTR_OBJ]), perm_bitmask);
	if (err != USP_ERR_OK) {
		USP_LOG_Info("Error creating rule[%s] errno(%d)", blobmsg_get_string(f_attr[ATTR_OBJ]), err);
	}
}

static int override_rule(ctrust_role_t role, struct blob_attr *rule)
{
	char *role_name;
	int rule_index;
	struct blob_attr *cur;
	int rem;
	enum {
		ATTR_NAME,
		ATTR_PERM,
		__MAX
	};
	struct blob_attr *f_attr[__MAX] = {NULL, NULL};
	const struct blobmsg_policy fp[__MAX] = {
		{ "name", BLOBMSG_TYPE_STRING},
		{"permission", BLOBMSG_TYPE_ARRAY}
	};

	if (blobmsg_parse(fp, __MAX, f_attr, blobmsg_data(rule), blobmsg_len(rule))) {
		USP_LOG_Info("Invalid role json, failed to parse");
		return INVALID;
	}

	if (!f_attr[ATTR_NAME]) {
		USP_LOG_Info("Invalid role json def, role name missing");
		return INVALID;
	}

	role_name = blobmsg_get_string(f_attr[ATTR_NAME]);
	if ((role == kCTrustRole_Untrusted) && (strcmp(role_name, "Untrusted") == 0)) {
		rule_index = kCTrustRole_Untrusted;
	} else if ((role == kCTrustRole_FullAccess) && (strcmp(role_name, "full_access") == 0)) {
		rule_index = kCTrustRole_FullAccess;
	} else {
		return INVALID;
	}

	USP_DM_RegisterRoleName(rule_index, role_name);
	blobmsg_for_each_attr(cur, f_attr[ATTR_PERM], rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_TABLE) {
			USP_LOG_Info("Invalid permission definition, skipping entry for %s", blobmsg_get_string(f_attr[ATTR_NAME]));
			continue;
		}

		create_rule_perm(rule_index, cur);
	}

	return USP_ERR_OK;
}

static int create_rule(int rule_index, struct blob_attr *rule)
{
	char *role_name;
	struct blob_attr *cur;
	int rem;
	enum {
		ATTR_NAME,
		ATTR_PERM,
		__ATTR_MAX
	};
	struct blob_attr *f_attr[__ATTR_MAX] = {NULL, NULL};
	const struct blobmsg_policy fp[__ATTR_MAX] = {
		{ "name", BLOBMSG_TYPE_STRING},
		{"permission", BLOBMSG_TYPE_ARRAY}
	};

	if (blobmsg_parse(fp, __ATTR_MAX, f_attr, blobmsg_data(rule), blobmsg_len(rule))) {
		USP_LOG_Info("Invalid role json, failed to parse");
		return INVALID;
	}

	if (!f_attr[ATTR_NAME]) {
		USP_LOG_Info("Invalid role json def, role name missing");
		return INVALID;
	}

	role_name = blobmsg_get_string(f_attr[ATTR_NAME]);
	if (strcmp(role_name, "Untrusted") == 0) {
		return INVALID;
	}
	if (strcmp(role_name, "full_access") == 0) {
		return INVALID;
	}

	if (rule_index > kCTrustRole_Max || rule_index <= kCTrustRole_Untrusted) {
		USP_LOG_Debug("Invalid rule index(%d)", rule_index);
		return INVALID;
	}

	USP_DM_RegisterRoleName(rule_index, role_name);
	blobmsg_for_each_attr(cur, f_attr[ATTR_PERM], rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_TABLE) {
			USP_LOG_Info("Invalid permission definition, skipping entry for %s", blobmsg_get_string(f_attr[ATTR_NAME]));
			continue;
		}

		create_rule_perm(rule_index, cur);
	}

	return rule_index;
}

static int create_default_roles(ctrust_role_t role)
{
	int err = USP_ERR_OK;

	if (role == kCTrustRole_FullAccess) {
		// Add full access role
		err |= USP_DM_RegisterRoleName(kCTrustRole_FullAccess, "full_access");
		err |= USP_DM_AddControllerTrustPermission(kCTrustRole_FullAccess, "Device.", PERMIT_ALL);
	} else if (role == kCTrustRole_Untrusted) {
		// Add Untrusted role
		err |= USP_DM_RegisterRoleName(kCTrustRole_Untrusted,  "Untrusted");
		err |= USP_DM_AddControllerTrustPermission(kCTrustRole_Untrusted, "Device.", PERMIT_NONE);
		err |= USP_DM_AddControllerTrustPermission(kCTrustRole_Untrusted, "Device.DeviceInfo.", PERMIT_GET | PERMIT_OBJ_INFO);
		err |= USP_DM_AddControllerTrustPermission(kCTrustRole_Untrusted, "Device.LocalAgent.ControllerTrust.Challenge.", PERMIT_GET_INST);
		err |= USP_DM_AddControllerTrustPermission(kCTrustRole_Untrusted, "Device.LocalAgent.ControllerTrust.Challenge.{i}.Alias", PERMIT_GET);
		err |= USP_DM_AddControllerTrustPermission(kCTrustRole_Untrusted, "Device.LocalAgent.ControllerTrust.Challenge.{i}.Description", PERMIT_GET);
		err |= USP_DM_AddControllerTrustPermission(kCTrustRole_Untrusted, "Device.LocalAgent.ControllerTrust.RequestChallenge()", PERMIT_GET | PERMIT_OPER);
		err |= USP_DM_AddControllerTrustPermission(kCTrustRole_Untrusted, "Device.LocalAgent.ControllerTrust.ChallengeResponse()", PERMIT_GET | PERMIT_OPER);
	} else {
		USP_LOG_Error("Invalid default role");
		err = USP_ERR_INTERNAL_ERROR;
	}

	return err;
}

static int create_rules_from_json(char *filename)
{
	int current_role_index = 0;
	bool is_full = false;
	bool is_untrusted = false;
	struct blob_buf bb;
	struct blob_attr *cur;
	int rem;
	enum {
		ROLES,
		__MAX
	};
	struct blob_attr *f_attr[__MAX] = {NULL};
	const struct blobmsg_policy fp[__MAX] = {
		{ "roles", BLOBMSG_TYPE_ARRAY }
	};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	if (blobmsg_add_json_from_file(&bb, filename) == false) {
		USP_LOG_Error("Failed to read roles from [%s] json file", filename);
		goto exit;
	}

	blobmsg_parse(fp, __MAX, f_attr, blobmsg_data(bb.head), blobmsg_len(bb.head));
	if (!f_attr[ROLES]) {
		USP_LOG_Error("Roles not defined in json file");
		goto exit;
	}

	// check if default roles are present in json to override with
	blobmsg_for_each_attr(cur, f_attr[ROLES], rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_TABLE)
			continue;

		if (override_rule(kCTrustRole_FullAccess, cur) == USP_ERR_OK) {
			is_full = true;
			break;
		}
	}
	if (is_full == false) {
		create_default_roles(kCTrustRole_FullAccess);
		is_full = true;
	}

	blobmsg_for_each_attr(cur, f_attr[ROLES], rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_TABLE)
			continue;

		if (override_rule(kCTrustRole_Untrusted, cur) == USP_ERR_OK) {
			is_untrusted = true;
			break;
		}
	}
	if (is_untrusted == false) {
		create_default_roles(kCTrustRole_Untrusted);
		is_untrusted = true;
	}

	current_role_index = kCTrustRole_Untrusted + 1;
	blobmsg_for_each_attr(cur, f_attr[ROLES], rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_TABLE)
			continue;

		if (current_role_index >= kCTrustRole_Max)
			break;

		if (create_rule(current_role_index, cur) != INVALID) {
			current_role_index++;
		}
	}

exit:
	if (is_full == false)
		create_default_roles(kCTrustRole_FullAccess);

	if (is_untrusted == false)
		create_default_roles(kCTrustRole_Untrusted);

	blob_buf_free(&bb);
	return USP_ERR_OK;
}

static int vendor_register_roles(void)
{
	int err = USP_ERR_OK;
	char role_json_file[256] = {0};
	char uci_role_path[] = "obuspa.global.role_file";


	get_value_from_uci(uci_role_path, role_json_file, 256);
	// Read rest of the roles from JSON, if defined
	if (role_json_file[0] == '\0') {
		err = create_default_roles(kCTrustRole_FullAccess);
		err |= create_default_roles(kCTrustRole_Untrusted);
	} else {
		err = create_rules_from_json(role_json_file);
	}

	if (err != USP_ERR_OK)
	{
		USP_LOG_Error("%s() failed", __FUNCTION__);
		return USP_ERR_INTERNAL_ERROR;
	}

	return USP_ERR_OK;
}

static int vendor_hook_init(void)
{
	vendor_hook_cb_t callbacks;

	memset(&callbacks, 0, sizeof(callbacks));
	callbacks.reboot_cb = reboot;
	callbacks.factory_reset_cb = factory_reset;
	callbacks.start_trans_cb = uspd_tran_start;
	callbacks.commit_trans_cb = uspd_tran_commit;
	callbacks.abort_trans_cb = uspd_tran_abort;
	callbacks.register_controller_trust_cb = vendor_register_roles;

	return USP_REGISTER_CoreVendorHooks(&callbacks);
}

static void dont_cached_params(str_vector_t *cached_vec)
{
	char cached_json_file[256] = {0};
	char uci_cpath[] = "obuspa.global.dm_caching_exclude";
	struct blob_buf bb;
	struct blob_attr *cur;
	int rem;
	enum {
		CACHED_PARAMS,
		__CACHED_MAX
	};
	struct blob_attr *f_attr[__CACHED_MAX] = {NULL};
	const struct blobmsg_policy fp[__CACHED_MAX] = {
		{ "dmcaching_exclude", BLOBMSG_TYPE_ARRAY }
	};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);
	STR_VECTOR_Init(cached_vec);

	if (get_value_from_uci(uci_cpath, cached_json_file, 256) != 0)
		goto exit;

	if (blobmsg_add_json_from_file(&bb, cached_json_file) == false)
		goto exit;

	blobmsg_parse(fp, __CACHED_MAX, f_attr, blobmsg_data(bb.head), blobmsg_len(bb.head));
	if (!f_attr[CACHED_PARAMS])
		goto exit;

	blobmsg_for_each_attr(cur, f_attr[CACHED_PARAMS], rem) {
		char *str = NULL;
		size_t len = 0;

		if (blobmsg_type(cur) != BLOBMSG_TYPE_STRING)
			continue;

		str = blobmsg_get_string(cur);
		len = strlen(str);
		if (strncmp(str, ROOT_PATH, 7) == 0 && str[len - 1] == '.') // Safety check
			STR_VECTOR_Add(cached_vec, str);
		else
			USP_LOG_Error("[%s:%d] Invalid entry[%s] in Cached json", __func__, __LINE__, str);
	}

exit:
	blob_buf_free(&bb);
}

// Function to calculate time required for get of complete dm and update the
// caching timer interval for optimal performance
static void update_adaptive_cached_time()
{
	struct blob_buf bb;
	uint32_t start, stop, delta;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	start = tu_uptime_msecs();
	blobmsg_add_string(&bb, "path", ROOT_PATH);
	uspd_call_inline_nocb("get", &bb);
	blob_buf_free(&bb);
	stop = tu_uptime_msecs();

	delta = (stop - start);
	g_uspd.cached_interval = MAX_CACHE_TIME + delta / 1000;
	g_uspd.ipc_timeout = DEFAULT_IPC_TIMEOUT + delta;
	USP_LOG_Debug("Max IPC timeout %d, cached %d", g_uspd.ipc_timeout, g_uspd.cached_interval);
}

static int uspd_global_init()
{
	memset(&g_uspd, 0, sizeof(struct uspd_global));

	g_uspd.ubus_ctx = ubus_connect(NULL);
	if (!g_uspd.ubus_ctx) {
		USP_LOG_Error("[%s:%d] ubus_connect failed", __func__, __LINE__);
		return USP_ERR_INTERNAL_ERROR;
	}

	// Required in init
	STR_VECTOR_Init(&g_uspd.group_vec);
	KV_VECTOR_Init(&g_uspd.cached_kv);

	g_uspd.mq_sockets[0] = -1;
	g_uspd.mq_sockets[1] = -1;
	g_uspd.cached_index = 0;
	g_uspd.cached_interval = MAX_CACHE_TIME;
	g_uspd.ipc_timeout = DEFAULT_IPC_TIMEOUT;
	g_uspd.trans_id = 0;
	g_uspd.last_gid = -1;
	g_uspd.op_state = USPD_INIT;

	pthread_cond_init(&g_uspd.ubus_cond, NULL);
	pthread_mutex_init(&g_uspd.ubus_lock, NULL);

	return USP_ERR_OK;
}

static void manual_vec_free(str_vector_t *vec)
{
	int i;

	for (i = 0; i < vec->num_entries; i++) {
		free(vec->vector[i]);
	}

	if (vec->vector)
		free(vec->vector);

	STR_VECTOR_Init(vec);
}

static int uspd_global_free()
{
	if (!g_uspd.ubus_ctx) {
		ubus_free(g_uspd.ubus_ctx);
		g_uspd.ubus_ctx = NULL;
	}

	g_uspd.mq_sockets[0] = -1;
	g_uspd.mq_sockets[1] = -1;
	g_uspd.cached_index = 0;
	g_uspd.cached_interval = MAX_CACHE_TIME;
	g_uspd.trans_id = 0;

	STR_VECTOR_Destroy(&g_uspd.cached_exclude_vec);
	KV_VECTOR_Destroy(&g_uspd.cached_kv);

	// This needs to be manually freed as the memory
	// monitoring stars with start function, so any memory allocated in
	// init needs to be freed manually.
	manual_vec_free(&g_uspd.group_vec);

	pthread_mutex_destroy(&g_uspd.ubus_lock);
	pthread_cond_destroy(&g_uspd.ubus_cond);

	return USP_ERR_OK;
}

static bool g_usp_object_available = false;

static void lookup_event_cb(struct ubus_context *ctx, struct ubus_event_handler *ev,
			  const char *type, struct blob_attr *msg)
{
	static const struct blobmsg_policy policy = {
		"path", BLOBMSG_TYPE_STRING
	};
	struct blob_attr *attr;
	const char *path;

	if (strcmp(type, "ubus.object.add") != 0)
		return;

	blobmsg_parse(&policy, 1, &attr, blob_data(msg), blob_len(msg));
	if (!attr)
		return;

	path = blobmsg_data(attr);
	if (strcmp(path, USP_OBJECT_NAME) == 0) {
		g_usp_object_available = true;
		uloop_end();
	}
}

static void lookup_timeout_cb(struct uloop_timeout *timeout)
{
	uloop_end();
}

static int wait_for_usp_raw_object()
{
#define USP_RAW_WAIT_TIMEOUT 60

	struct ubus_context *uctx;
	int ret;
	uint32_t ubus_id;
	struct ubus_event_handler add_event;
	struct uloop_timeout u_timeout;

	g_usp_object_available = false;
	uctx = ubus_connect(NULL);
	if (uctx == NULL) {
		USP_LOG_Error("Can't create ubus context");
		return USP_ERR_INTERNAL_ERROR;
	}

	uloop_init();
	ubus_add_uloop(uctx);

	// register for add event
	memset(&add_event, 0, sizeof(struct ubus_event_handler));
	add_event.cb = lookup_event_cb;
	ubus_register_event_handler(uctx, &add_event, "ubus.object.add");

	// check if object already present
	ret = ubus_lookup_id(uctx, USP_OBJECT_NAME, &ubus_id);
	if (ret == 0) {
		g_usp_object_available = true;
		goto end;
	}

	// Set timeout to expire lookup
	memset(&u_timeout, 0, sizeof(struct uloop_timeout));
	u_timeout.cb = lookup_timeout_cb;
	uloop_timeout_set(&u_timeout, USP_RAW_WAIT_TIMEOUT * 1000);

	uloop_run();
	uloop_done();

end:
	ubus_free(uctx);

	if (g_usp_object_available == false) {
		USP_LOG_Error("### %s object not found", USP_OBJECT_NAME);
		return USP_ERR_INTERNAL_ERROR;
	}

	return USP_ERR_OK;
}

// Public methods
int vendor_uspd_init(void)
{
	int fault = USP_ERR_OK;

	fault = wait_for_usp_raw_object();
	if (fault) {
		return USP_ERR_INTERNAL_ERROR;
	}

	fault = uspd_global_init();
	if (fault) {
		return USP_ERR_INTERNAL_ERROR;
	}

	fault = uspd_register_schema();

	if (fault == USP_ERR_OK)
		fault = vendor_hook_init();

	return fault;
}

int vendor_uspd_start(void)
{
	int fault = USP_ERR_OK;


	// update exclude vector from json
	update_adaptive_cached_time();
	dont_cached_params(&g_uspd.cached_exclude_vec);

	fault = vendor_exec_init();
	if (fault == USP_ERR_OK)
		fault = OS_UTILS_CreateThread("UBUS_Thread", start_ubus_thread, NULL);

	return fault;
}

int vendor_uspd_stop(void)
{
	uspd_global_free();

	return USP_ERR_OK;
}
